/* ========================== C MeatAxe =============================
   charpol.c -  Characteristic polynomial of a matrix

   (C) Copyright 1994 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@tiffy.math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ================================================================== */


/* $Id: charpol.c,v 1.4 1994/07/07 05:52:21 mringe Exp $
 *
 * $Log: charpol.c,v $
 * Revision 1.4  1994/07/07  05:52:21  mringe
 * Compiler warning
 *
 * Revision 1.3  1994/07/04  11:47:36  mringe
 * Dimensionlimit bei charpol()
 *
 * Revision 1.2  1994/05/04  11:42:29  mringe
 * hmmm.
 *
 * Revision 1.2  1994/05/04  11:42:29  mringe
 * hmmm.
 *
 * Revision 1.1  1994/04/07  21:11:02  mringe
 * Initial revision
 *
 */

#include <string.h>
#include <stdlib.h>
#include "meataxe.h"




/* ------------------------------------------------------------------
   Function prototypes
   ------------------------------------------------------------------ */

static void cleanup _PL((void));
static int init _PL((matrix_t *matrix));
static void spinup_cyclic _PL((void));
static poly_t *mkpoly _PL((void));



/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

static long fl, nor;		/* Field and size */
static long *piv = NULL;	/* Pivot table */
static char *ispiv = NULL;	/* Pivot flags */
static PTR mat = NULL,		/* The matrix */
    A = NULL,			/* Work space (for spin-up) */
    B = NULL;			/* Work space II (coefficients) */
static long dim;		/* Dimension reached so far */
static long n;			/* Dimension of cyclic subspace */

/* ------------------------------------------------------------------
   cleanup() - Free memory
   ------------------------------------------------------------------ */

static void cleanup()

{
    if (mat != NULL) free(mat);
    if (A != NULL) free(A);
    if (B != NULL) free(B);
    if (piv != NULL) free(piv);
    if (ispiv != NULL) free(ispiv);
    mat = A = B = NULL;
    piv = NULL;
    ispiv = NULL;
}


/* ------------------------------------------------------------------
   init()
   ------------------------------------------------------------------ */

static int init(matrix)
matrix_t *matrix;

{	
    /* Set some global variables
       ------------------------- */
    if (matrix->nor != matrix->noc) MTXFAIL(ERR_NOTSQUARE,-1);
    fl = matrix->fl;
    nor = matrix->nor;
    dim = 0;

    /* Allocate memory
       --------------- */
    cleanup();
    zsetlen(fl,nor);
    mat = zalloc(nor);
    A = zalloc(nor+1);
    B = zalloc(nor+1);
    piv = (long *) malloc((size_t)(nor+2)*sizeof(long));
    ispiv = (char *) malloc((size_t)(nor+2));
    if (piv == NULL || ispiv == NULL) MTXFAIL(ERR_NOMEM,-1);

    /* Initialize memory
       ----------------- */
    memcpy(mat,matrix->d,zsize(nor));
    memset(ispiv,0,(size_t)(nor+2));

    return 0;
}


/* ------------------------------------------------------------------
   mkpoly() - Make polynomial for the latest cyclic subspace
   ------------------------------------------------------------------ */

static poly_t *mkpoly()

{
    int k;
    PTR x;
    poly_t *pol;
    
    pol = polalloc(fl,nor);
    x = B;
    zadvance(&x,n);
    pol->deg = n;
    for (k = 1; k <= n; ++k)
	pol->buf[k-1] = zextract(x,k);
    pol->buf[n] = F_ONE;
    return pol;
}


/* ------------------------------------------------------------------
   spinup_cyclic() - Spin up one cyclic subaspace
   ------------------------------------------------------------------ */

static void spinup_cyclic()

{   PTR a, b;
    long pv, k;
    FEL f;

/*fprintf(stderr,"spinup_cyclic(): dim=%ld\n",dim);*/

    a = A;
    zadvance(&a,dim);
    b = B;
    zmulrow(b,F_ZERO);
    n = 0;
    while ((pv = zfindpiv(a,&f)) != 0)
    {	
	PTR x, y;

    	/* Add new vector to basis
	   ----------------------- */
	piv[dim+1+n] = pv;
/*fprintf(stderr,"piv[%ld]=%ld\n",dim+1+n,pv);*/
	ispiv[pv] = 1;
	++n;
	zinsert(b,n,F_ONE);

	/* Calculate the next vector
	   ------------------------- */
	x = a;
	zadvance(&a,(long)1);
	zmaprow(x,mat,nor,a);
/*
{long ii; fprintf(stderr," x ="); for (ii=1; ii < nor; ++ii)
 fprintf(stderr,"%ld",zftoi(zextract(x,ii))); fprintf(stderr,"\n xA=");
 for (ii=1; ii < nor; ++ii) fprintf(stderr,"%ld",zftoi(zextract(a,ii)));
 fprintf(stderr,"\n");
}
*/
	y = b;
	zadvance(&b,(long)1);
	zmulrow(b,F_ZERO);
	for (k = 1; k < nor; ++k)
	    zinsert(b,k+1,zextract(y,k));
    
	/* Clean with existing basis vectors
	   --------------------------------- */
	x = A;
	y = B;
	for (k = 1; k <= dim+n; ++k)
	{   f = zdiv(zextract(a,piv[k]),zextract(x,piv[k]));
	    zaddmulrow(a,x,zsub(F_ZERO,f));
	    if (k > dim)
	    {	zaddmulrow(b,y,zsub(F_ZERO,f));
		zadvance(&y,(long)1);
	    }
	    zadvance(&x,(long)1);
	}
    }
    dim += n;
}


/* ------------------------------------------------------------------
   charpolfactor() - Returns next (reducible) factor
   ------------------------------------------------------------------ */

poly_t *charpolfactor(mat)
matrix_t *mat;

{
    PTR seed;
    int i;

    /* If called with mat != NULL, initialize everything
       ------------------------------------------------- */
    if (mat != NULL)
    	if (init(mat) != 0)
    	    return NULL;

    /* If there is nothing left to do, return NULL
       ------------------------------------------- */
    /* we could call cleanup() here... */
    if (dim >= nor) return NULL;

    /* Prepare the next seed vector
       ---------------------------- */
    zsetlen(fl,nor);
    seed = A;
    zadvance(&seed,dim);
    for (i = 1; i <= nor && ispiv[i] != 0; ++i);
    zmulrow(seed,F_ZERO);
    zinsert(seed,i,F_ONE);

    /* Spin up and return the polynomial 
       --------------------------------- */
    spinup_cyclic();
    return mkpoly();    
}



/* ------------------------------------------------------------------
   charpol() - Returns factorized characteristic polynomial. If
   deglimit > 0, only factors with degree <= deglimit are considered.
   ------------------------------------------------------------------ */

fpoly_t *charpol(mat,deglimit)
matrix_t *mat;
long deglimit;

{
    fpoly_t *cpol = fpolalloc();
    poly_t *p;
    poly_t *tt = NULL;
    poly_t *complete = NULL;

    if (deglimit > 0 && deglimit < 4)
    {
	long a = mat->fl;
	long d = deglimit;
	while (--d > 0) a *= mat->fl;
	if (a < 1) MTXFAIL(ERR_RANGE,NULL);
	tt = polalloc(mat->fl,a);
	tt->buf[1] = zsub(F_ZERO,F_ONE);
	complete = polalloc(mat->fl,0);
    }

    for (p = charpolfactor(mat); p != NULL; p = charpolfactor(NULL))
    {
	fpoly_t *fac;
	int i;
	if (tt != NULL)
	{
	    poly_t *tmp = p;
	    polmul(complete,p);
	    p = polgcd(p,tt);
	    polfree(tmp);
	}
	fac = factorization(p);
    	polfree(p);
	for (i = 0; i < fac->len; ++i)
	    if (deglimit <= 0 || fac->p[i]->deg <= deglimit)
	    	fpolmulp(cpol,fac->p[i],fac->e[i]);
	fpolfree(fac);
    }

    if (tt != NULL)
    {
	int i;

	/* Calculate the multiplicities
	   ---------------------------- */
	for (i = 0; i < cpol->len; ++i)
	{
	    poly_t *x = poldup(complete);
	    int k;
	    for (k = 0; ; ++k)
	    {
		poly_t *quot = poldivmod(x,cpol->p[i]);
		if (x->deg >= 0)
		{
		    polfree(x);
		    polfree(quot);
		    break;
		}
		else
		{
		    polfree(x);
		    x = quot;
		}
	    }
	    cpol->e[i] = k;
	}
	polfree(tt);
	polfree(complete);
    }
    return cpol;
}
