/* ========================== C MeatAxe =============================
   This program finds the peak words and does the generalized
   kondensation for each composition factor.

   (C) Copyright 1994 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@tiffy.math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ================================================================== */


/* $Id: pwkond.c,v 1.2 1994/07/07 09:40:33 mringe Exp $
 *
 * $Log: pwkond.c,v $
 * Revision 1.2  1994/07/07  09:40:33  mringe
 * *** empty log message ***
 *
 * Revision 1.1  1994/07/07  09:31:03  mringe
 * Initial revision, aus mkpeak.c und gkond.c zusammengesetzt.
 *
 */


#include <string.h>
#include <stdlib.h>

#include "meataxe.h"
#include "lattice.h"
#include "files.h"


/* ------------------------------------------------------------------
   Function prototypes
   ------------------------------------------------------------------ */

static void init2 _PL((void));
static void power _PL((int i));
static void gkond _PL((int i, matrix_t *k, matrix_t *w, char *name));
static void kond _PL((int cf));


/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

long dim;
long quotdim;
static matrix_t *gen[MAXGEN];		/* Generators */
static matrix_t *tmp1, *tmp2;
static basis_t basis;
int opt_G = 0;

static char *helptext[] = {
"SYNTAX",
"    pwkond [<Options>] <Name>",
"",
"OPTIONS",
"    -Q            Quiet, no messages.",
"    -V            Verbose, more messages.",
"    -G            GAP output (implies -Q).",
"    -T <Seconds>  Set CPU time limit",
"    -e <List>     Exclude words from search. <List> is a",
"                  comma-separated list of numbers or ranges (X-Y).",
"",
"FILES",
"    Input files are produced by CHOP (see `chop -help').",
"    Output is written to the following files:",
"       <Name><Dim><a>.<N>k        Kondensed generators",
"       <Name><Dim><a>.np          Kondensed peak word",
"       <Name><Dim><a>.im          Image used for kondenstion",
"       <Name><Dim><a>.k           Unkondense matrix (used by mkinc)",
NULL};

static proginfo_t pinfo =
   { "pwkond", "Peakword Kondensation",
     "$Revision: 1.2 $", helptext };



/* ------------------------------------------------------------------
   init()
   ------------------------------------------------------------------ */

static void init2()

{
    char fn[MAXBASENAME+10];
    int i;
    long fl;

    for (i = 0; i < ngen; ++i)
    {
	sprintf(fn,"%s.%d",cfbasename,i+1);
	gen[i] = matload(fn);
    }
    fl = gen[0]->fl;
    initbasis(&basis,ngen,gen);
    dim = gen[0]->nor;
    tmp1 = matalloc(fl,dim,dim);
    tmp2 = matalloc(fl,dim,dim);
}


/* ------------------------------------------------------------------
   power()
   ------------------------------------------------------------------ */

static void power(i)
int i;

{
    long nul1, nul2, k0;

    MESSAGE(0,("word=%ld, ",cfinfo[i].peakword));
    mkword(&basis,cfinfo[i].peakword);
    matmove(tmp1,basis.w);
    matmove(tmp2,tmp1);
    k0 = 1;
    nul2 = nullity(tmp2);
    nul1 = -1;
    while (nul2 != nul1)	/* while nullity not stable */
    {
	k0 <<= 1;
	nul1 = nul2;
	matmove(tmp2,tmp1);
	matmul(tmp2,tmp1);
	matmove(tmp1,tmp2);
	matmul(tmp1,basis.w);
	matmove(tmp2,tmp1);
	nul2 = nullity(tmp2);
    }
    MESSAGE(0,("pwr=%ld, nul=%ld, mult=%ld, ",k0,nul2,cfinfo[i].mult));
    fflush(stdout);

    /* Consistency check
       ----------------- */
    if (nul2 != cfinfo[i].mult * cfinfo[i].spl)
	FATAL("Something is wrong...");
    quotdim = nul2;
}



/* ------------------------------------------------------------------
   gkond()
   ------------------------------------------------------------------ */

static void gkond(i,k,w,name)
int i;
matrix_t *k, *w;
char *name;

{
    char fn[MAXBASENAME+10];
    matrix_t *x1, *x2;

    x1 = matdup(k);
    matmul(x1,w);

    x2 = matalloc(x1->fl,x1->nor,quotdim);
    zquot(x1->d,x1->nor,x2->d);

    sprintf(fn,"%s%s.%s",cfbasename,cfname(i),name);
    matsave(x2,fn);
    matfree(x1);
    matfree(x2);
}


/* ------------------------------------------------------------------
   kond() - Generalized kondensation for one irreducible
   ------------------------------------------------------------------ */

static void kond(cf)
int cf;

{
    char fn[MAXBASENAME+10];
    matrix_t *kern, *bild, *m, *k;
    int j;

    MESSAGE(0,("Kondensing %s%s: ",cfbasename,cfname(cf)));
    fflush(stdout);
		
    /* Make the peak word, find its stable power,
       calculate both kernel and image.
       ------------------------------------------- */
    power(cf);
    kern = nullsp(tmp1);
    bild = echelon(tmp1);
    zquotinit(bild->d,bild->nor,NULL);

    /* Write out the image
       ------------------- */
    sprintf(fn,"%s%s.im",cfbasename,cfname(cf));
    matsave(bild,fn);

    /* Write out the `unkondense matrix'
       --------------------------------- */
    m = matalloc(kern->fl,kern->nor,quotdim);
    zquot(kern->d,kern->nor,m->d);
    k = matinv(m);
    matmul(k,kern);
    sprintf(fn,"%s%s.k",cfbasename,cfname(cf));
    matsave(k,fn);

    /* Kondense all generators
       ----------------------- */
    MESSAGE(0,("("));
    for (j = 0; j < ngen; ++j)
    {
	sprintf(fn,"%dk",j+1);
	gkond(cf,k,gen[j],fn);
	MESSAGE(0,(" %d",j+1));
    }
    MESSAGE(0,(" )\n"));

    /* Kondense the peak word
       ---------------------- */
    gkond(cf,k,basis.w,"np");

    matfree(m); matfree(k);
    matfree(kern); matfree(bild);
}




#define MAXLOCK 100
basis_t cfbasis[MAXCF];
matrix_t *tmp[MAXCF];
long exclude[MAXLOCK][2];
int nexclude=0;



static int isexcluded(w)
long w;

{	int i;

	for (i = 0; i < nexclude; ++i)
		if (w >= exclude[i][0] && w <= exclude[i][1])
			return 1;
	return 0;
}

static void parselist(c)
char *c;

{	long a, b;

	while (*c != 0)
	{	a = b = 0;
		while (*c >= '0' && *c <= '9')
			a = a * 10 + (*c++ - '0');
		if (*c == '-')
		{	++c;
			while (*c >= '0' && *c <= '9')
				b = b * 10 + (*c++ - '0');
		}
		else
			b = a;
		if (a == 0 || b == 0 || a > b) 
			FATAL("BAD ARGUMENTS");
		exclude[nexclude][0] = a;
		exclude[nexclude][1] = b;
		++nexclude;
		if (*c == ',') ++c;
	}
}


static void init(argc, argv)
int argc;
char *argv[];

{
    char fn[50];
    int i;
    matrix_t *cfgen[MAXGEN];

    readcfinfo();

    /* Read the generators for each composition factor
       ----------------------------------------------- */
    for (i = 0; i < ncf; ++i)
    {
	int k;
	for (k = 0; k < ngen; ++k)
	{
	    sprintf(fn,"%s%s.%d",cfbasename,cfname(i),k+1);
	    MESSAGE(1,("Reading %s\n",fn));
	    cfgen[k] = matload(fn);
	}
	initbasis(cfbasis+i,ngen,cfgen);
	tmp[i] = matalloc(cfgen[0]->fl,cfinfo[i].dim,cfinfo[i].dim);
	for (k = 0; k < ngen; ++k)
	    matfree(cfgen[k]);
	cfinfo[i].peakword = -1;
    }
}


/* ------------------------------------------------------------------
   finished() - Find out if a peak word has been found for each
	module.
   ------------------------------------------------------------------ */

static int finished()

{
    int i;

    for (i = 0; i < ncf; ++i)
	if (cfinfo[i].peakword == -1) return 0;
    return 1;
}


/* ------------------------------------------------------------------
   try() - Try another word
   ------------------------------------------------------------------ */

static int try(w)
long w;

{
    int i, ppos = -1;
    long nul;

    if (isexcluded(w)) return -1;
    MESSAGE(1,("Word %ld:",w));
    for (i = 0; i < ncf; ++i)  /* For each composition factor... */
    {
	mkword(cfbasis+i,w);
	matmove(tmp[i],cfbasis[i].w);
	nul = nullity(tmp[i]);
        MESSAGE(1,(" %ld",nul));
	if (nul != 0 && nul != cfinfo[i].spl)
	    return -1;
	if (nul == cfinfo[i].spl)
	{
	    if (ppos >= 0 || cfinfo[i].peakword > 0)
		return -1;  /* Nullity should be 0 */
	    matmove(tmp[i],cfbasis[i].w);
	    matmul(tmp[i],cfbasis[i].w);
	    nul = nullity(tmp[i]);
	    if (nul != cfinfo[i].spl)
		return -1;  /* Nullity is not stable */
	    ppos = i;
	}
    }
    if (ppos > -1) /* we have found a new peak word */
	cfinfo[ppos].peakword = w;
    return ppos;
}



/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */


int main(argc, argv)
int argc;
char *argv[];

{
    int i;
    long w;

    /* Parse command line
       ------------------ */
    mtxinit();
    initargs(argc, argv, &pinfo);
    while ((i = zgetopt("Ge:")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': opt_G = 1; msg_level = -100; break;
	    case 'e':
		parselist(opt_text);
		break;
	}
    }
    if (opt_ind != argc-1) errexit(ERR_BADUSAGE,pinfo.name);
    setbasename(argv[argc-1]);
    MESSAGE(0,("\n*** PEAK WORD KONDENSATION ***\n\n"));
    init(argc,argv);
    init2();

    for (w = 1; !finished(); w = nextword(w))
    {
	i = try(w);
	MESSAGE(1,("\n"));
	if (i >= 0)
	{
	    MESSAGE(0,("Peak word for %s%s is %ld, nullity = %ld\n",
		cfbasename,cfname(i),cfinfo[i].peakword,
		cfinfo[i].spl));
	    fflush(stdout);
	    kond(i);
	}
    }
    writecfinfo();
    if (opt_G)
    {
	printf("MeatAxe.PeakWords := [");
	for (i = 0; i < ncf; ++i)
	{
	    if (i > 0) printf(",");
	    printf("%ld",cfinfo[i].peakword);
	}
	printf("];\n");
    }
    if (msg_level >= 0)
    {
	printf("\n");
	prtimes();
    }
    return 0;
}


